@@ -518,7 +518,7 @@ handle_input({frame_header_1_0, Mode},
     end,
     case Size of
         8 -> % length inclusive
-            {State, {frame_header_1_0, Mode}, 8}; %% heartbeat
+            State; %% heartbeat
         _ ->
             switch_callback(State, {frame_payload_1_0, Mode, DOff, Channel}, Size - 8)
     end;
@@ -545,8 +545,9 @@ start_1_0_connection(sasl, State = #v1{sock = Sock}) ->
     Ms = {array, symbol,
           case application:get_env(rabbitmq_amqp1_0, default_user)  of
               {ok, none} -> [];
-              {ok, _}    -> ["ANONYMOUS"]
-          end ++ [ atom_to_list(M) || M <- auth_mechanisms(Sock)]},
+              {ok, _}    -> [<<"ANONYMOUS">>]
+          end ++
+              [list_to_binary(atom_to_list(M)) || M <- auth_mechanisms(Sock)]},
     Mechanisms = #'v1_0.sasl_mechanisms'{sasl_server_mechanisms = Ms},
     ok = send_on_channel0(Sock, Mechanisms, rabbit_amqp1_0_sasl),
     start_1_0_connection0(sasl, State);
