@@ -33,7 +33,9 @@
      {mfa,         {rabbit_registry, register,
                     [exchange, <<"x-consistent-hash">>, ?MODULE]}},
      {requires,    rabbit_registry},
-     {enables,     kernel_ready}]}).
+     {enables,     kernel_ready},
+     {cleanup,     {rabbit_registry, unregister,
+                    [exchange, <<"x-consistent-hash">>]}}]}).
 
 -rabbit_boot_step(
    {rabbit_exchange_type_consistent_hash_mnesia,
