@@ -11,7 +11,7 @@
 %%   The Original Code is RabbitMQ Management Plugin.
 %%
 %%   The Initial Developer of the Original Code is GoPivotal, Inc.
-%%   Copyright (c) 2010-2014 GoPivotal, Inc.  All rights reserved.
+%%   Copyright (c) 2010-2015 Pivotal Software, Inc.  All rights reserved.
 %%
 
 -module(rabbit_mgmt_format).
@@ -157,9 +157,8 @@ internal_user(User) ->
      {tags,          tags(User#internal_user.tags)}].
 
 user(User) ->
-    [{name,         User#user.username},
-     {tags,         tags(User#user.tags)},
-     {auth_backend, User#user.auth_backend}].
+    [{name, User#user.username},
+     {tags, tags(User#user.tags)}].
 
 tags(Tags) ->
     list_to_binary(string:join([atom_to_list(T) || T <- Tags], ",")).
