@@ -11,7 +11,7 @@
 %% The Original Code is RabbitMQ Federation.
 %%
 %% The Initial Developer of the Original Code is GoPivotal, Inc.
-%% Copyright (c) 2007-2014 GoPivotal, Inc.  All rights reserved.
+%% Copyright (c) 2007-2015 Pivotal Software, Inc.  All rights reserved.
 %%
 
 -module(rabbit_shovel_test_dyn).
@@ -257,7 +257,7 @@ invalid_param(Value) -> invalid_param(Value, none).
 valid_param(Value) -> valid_param(Value, none).
 
 lookup_user(Name) ->
-    {ok, User} = rabbit_auth_backend_internal:check_user_login(Name, []),
+    {ok, User} = rabbit_access_control:check_user_login(Name, []),
     User.
 
 clear_param(Name) ->
